/* N-body calculation, adapted from C* Ref. Man., pg. 38 ff. */

#include <stdio.hs>
#include <math.hs>

/* Number of bodies in the calculation */
#define N 6
/* Largest power of 2 < N-1 */
#define N2 4

float G;     /* Gravitational constant */

/* The bodies */

domain particle { float x, y, z, m, vx, vy, vz, ax, ay, az; };

domain particle B[N];

/* Pairs of particles */

domain pair { int i, j, jx; };

domain pair P[N][N - 1];

/* Pairs will be represented symmetrically.  The function setup
   (below) initializes these so that 0 <= i,j < N, i != j. */

void pair::setup()
 { int offset;
   offset = (this - &P[0][0]);
   i = offset / (N-1);  jx = j = offset % (N-1);
   if ( j >= i ) j++;
 }

void pair::print_pairs()
 { mono int k, l;
   for ( k = 0; k < N; k++ )
    { printf("%2d: ", k);
      for ( l = 0; l < (N - 1); l++ ) printf("%2d ", P[k][l].j);
      printf("\n"); }
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */

void init()
 { int k;  float a, b, c;
   for ( k = 0; k < N; k++ )
    { fscanf(datap, "%e%e%e", &a, &b, &c);
      B[k].x = a;  B[k].y = b;  B[k].z = c;
    };
   for ( k = 0; k < N; k++ )
    { fscanf(datap, "%e%e%e", &a, &b, &c);
      B[k].vx = a;  B[k].vy = b;  B[k].vz = c;
    };
   for ( k = 0; k < N; k++ ) { fscanf(datap, "%e", &a);  B[k].m = a; };
   fscanf(datap, "%e", &G);
 }

void particle::print_state()
 { mono int k;
   printf(
  "\n        x         y         z         vx        vy        vz        m\n");
   for ( k = 0; k < N; k++ )
     printf("%3d %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e %9.2e\n",
              k, B[k].x, B[k].y, B[k].z, B[k].vx, B[k].vy, B[k].vz, B[k].m);
 }

void particle::print_ac()
 { mono int k;
   printf("\n        ax        ay        az\n");
   for ( k = 0; k < N; k++ )
     printf("%3d %9.2e %9.2e %9.2e\n", k, B[k].ax, B[k].ay, B[k].az);
 }

/* Calculate accelerations */

void accel()
 { [domain particle].{ ax = ay = az = 0.0; };
   [domain pair].
    { float dx, dy, dz, r2, r3, q;
      mono int k, l;
      dx = B[j].x - B[i].x;  dy = B[j].y - B[i].y;  dz = B[j].z - B[i].z;
      r2 = dx*dx + dy*dy + dz*dz;  r3 = r2*sqrt(r2);
      q = G*(B[j].m)/r3;
      /* Individual acceleration terms */
      dx *= q;  dy *= q;  dz *= q;
      /* Sum acceleration terms in time log N */
      if ( jx < (N - 1 - N2) )
       { dx += (this + N2)->dx;
	 dy += (this + N2)->dy;
	 dz += (this + N2)->dz;
       };
      for ( k = (N2 >> 1); k >= 1; k >>= 1)
       if ( jx < k )
        { dx += (this + k)->dx;
	  dy += (this + k)->dy;
	  dz += (this + k)->dz;
        };
      if (jx == 0)
       { B[i].ax ,= dx;  B[i].ay ,= dy;  B[i].az ,= dz; };
    };
 }


/* Modified version of Euler's method.  Performs ns steps with
   time increment dt. */

void euler(ns, dt)
  int ns;  float dt;
 { int k;
   for (k = 0; k < ns; k++)
    { accel();
      [domain particle].
       { float w;
         w = vx + ax*dt;  x += 0.5*dt*(vx + w); vx = w;
         w = vy + ay*dt;  y += 0.5*dt*(vy + w); vy = w;
         w = vz + az*dt;  z += 0.5*dt*(vz + w); vz = w;
       };
    };
 }

void main()
 { int ns, np, k;  float dt;
   /* Performs np integrations, each ns steps of dt */
   [domain pair].{ setup(); };
   datap = fopen("sixbody.dat", "r");  init();
   printf("\n   G = %13.6e\n", G);
   printf("\n time = 0");
   [domain particle].{ print_state(); };
   fscanf(datap, "%d%d%e", &ns, &np, &dt);
   for ( k = 0; k < np; k++ )
    { euler(ns, dt);
      printf("\n time = %9.3f\n", (k + 1)*ns*dt);
      [domain particle].{ print_state(); };
    };
 }

/* Arrangement of data file:

     x      y      z
     .      .      .     Initial positions
     .      .      .     for N points
     .      .      .

    vx     vy     vz
     .      .      .     Initial velocities
     .      .      .
     .      .      .

     m
     .                   Masses
     .
     .

     G                   Gravitational constant

    ns     np     dt     Integration parameters  */

